#include <cctype>
#include <cstdarg>
#include <ctime>
#include <errno.h>
#include <fcntl.h>
#include <iostream>
#include <unistd.h>
#include <sys/stat.h>
#include <fstream>

#define TELOPTS
#define TELCMDS
#include <arpa/telnet.h>

#include "playertask.hh"
#include "taskmanager.hh"
#include "stringfun.hh"
#include "events.hh"

////////////

static const bool BROKEN_TERM = true;
static const bool SUPPORT_2x  = false;
static const bool COLOURS     = true;

class playerdata
{
	playertask &owner;
	mutable unsigned char attr;	
    void colour(unsigned char newattr) const
    {
    	if(newattr == attr || !COLOURS)return;
    	
    	string Buf = "\33[";
    	
    	if(BROKEN_TERM || newattr != 7)
    	{
    		static const char Swap[] = "04261537";
    		bool pp;
    		#define P(a)    do{if(pp)Buf+=';';pp=true;Buf+=a;}while(0)
    		#define P2(a,b) do{if(pp)Buf+=';';pp=true;Buf+=a;Buf+=b;}while(0)
    		if(SUPPORT_2x)
    		{
    			if((attr & 0x88) && !(newattr & 0x88)) P('0');
    			else if((attr & 0x80) && !(newattr & 0x80)) P("25");
    			else if((attr & 0x08) && !(newattr & 0x08)) P("21");
    			else if((newattr & 0x80) && !(attr & 0x80)) P('5');
    			else if((newattr & 0x08) && !(attr & 0x08)) P('1');
    		}
    		else
    		{
    			if(((attr & 0x80) && !(newattr & 0x80))
    			|| ((attr & 0x08) && !(newattr & 0x08)))
    			{
    				P("0");
    				attr = BROKEN_TERM ? (attr&0x77) ^ 0x77 : 7;
    			}
    			if((newattr & 0x80) && !(attr & 0x80)) P('5');
    			if((newattr & 0x08) && !(attr & 0x08)) P('1');
    		}
    		if((newattr & 0x70) != (attr & 0x70)) P2('4', Swap[(newattr>>4)&7]);
    		if((newattr & 0x07) != (attr & 0x07)) P2('3', Swap[newattr&7]);
    		#undef P
    	}
    	Buf += 'm';
    	owner.echo(Buf, false);
   		attr = newattr;
    }
    
    void displayprompt()
    {
    	litecho(prompt);
    	save_ok = true;
    }
    
    inline string playerfilename() const
    {
    	return playerfilename(name);
    }
    string playerfilename(const string &whom) const
    {
    	return "players/" + whom;
    }

public:
	enum situation
	{
		// Never add new situations in the
		// middle, always add them to the end!
		S_LOGIN = 0,
		S_GAME,
		S_LOGINPASS,
		S_CREATE,
		S_PASSCHECK
	} situation;
	
	string name;
	string pass;
	string prompt;
	time_t creation;
	
	bool save_ok;
	
public:
	playerdata(playertask &o) : owner(o), attr(0x07), situation(S_LOGIN), save_ok(false)
	{
    	litecho(
    	    "`F---------------------------------------------------------------------------\n"
    	    "`A*** `FDUD4 `Aruntime - copyright (C) `21992`A,`22000`A Bisqwit (`Bhttp://iki.fi/bisqwit/`A)\n"
    	    "`F---------------------------------------------------------------------------\n"
    	    "`3DUD login v"VERSION" - recommended to have square font.\n"
    	    "`7\n"
    	    "Enter your character name (one word, case insignificant): ");
	}
	
	~playerdata()
	{
		save();
	}
	
	void load()
	{
		load(name);
	}
	
	void load(string whom)
	{
		ifstream f(playerfilename(whom).c_str());
		if(f.good())
		{
			string line;
			while(f.good())
			{
				getline(f, line);
				vector<string> args;
				string tmp = strtoupper(argsplit(args, line));
				if(tmp == "SITUATION")
				{
					situation = (enum situation)strtol(argsplit(args, line).c_str(), NULL, 10);
					continue;
				}
				if(tmp == "CREATION")
				{
					creation = (time_t)strtod(argsplit(args, line).c_str(), NULL);
					continue;
				}
				if(tmp == "PROMPT")  { prompt = line; continue; }
				if(tmp == "PASS")    { pass   = line; continue; }
			}
		}
		else
		{
			situation = S_CREATE;
			litecho("\n"
				"`FYou do not exist. `7Make a password for you, and we will fix it.\n"
				"Here, please: ");
			noecho();
		}
	}
	
	void save()
	{
		if(save_ok)
		{
			ofstream f(playerfilename().c_str());
			f << "SITUATION " << (int)situation << "\n"
			     "PROMPT " << prompt << "\n"
			     "PASS " << pass << "\n"
			     "CREATION " << creation << "\n"
			;
		}
	}

    void telnetsend(int dummy, ...) const
    {
    	va_list ap;
    	va_start(ap, dummy);
    	
    	int a;
    	string Buf;
    	
    	Buf += (char)dummy;
    	
    	while((a = va_arg(ap, int)) >= 0)
    		Buf += (char)a;
    		
    	owner.echo(Buf, false);
    }

    void doecho() const
    {
    	// Client will when we won't.
    	telnetsend(IAC, WONT, TELOPT_ECHO, -1);
    }

    void noecho() const
    {
    	// Client won't when we will
    	telnetsend(IAC, WILL, TELOPT_ECHO, -1);
    }
    
    void caninput()
    {
		telnetsend(IAC, GA, -1);
		colour(7);
    }
    
    void litecho(const string &s) const
    {
    	const char *q = s.c_str();
    	const char *p;
    	
    	for(;;)
    	{
    		const char *d = q;
    	retest:
    		p = strchr(d, '`');
    		if(!p)
    		{
    			owner.echo(q, false);
    			return;
    		}
    		owner.echo(string(q, 0, p-q), false);
    		if(!p[1])break;
    		if(p[1] == '`')
    			owner.echo("`", false);
    		else if((p[1] >= '0' && p[1] <= '9'))
    			colour((attr&0xF0) | (p[1]-'0'));
    		else if((p[1] >= 'A' && p[1] <= 'F'))
    			colour((attr&0xF0) | (p[1]-'A'+10));
    		else
    		{
    			d = p+1;
    			goto retest;
    		}
    		q = p + 2;
    	}
    }
    
    void handleinput()
	{
		vector<string> args;
		string input = owner.input;
		
		switch(situation)
		{
			case S_LOGIN:
				name = strtolower(argsplit(args, input));
				load();
				if(situation != S_CREATE)
				{
					situation = S_LOGINPASS;
					litecho("Password: ");
					noecho();
				}
				break;
			case S_CREATE:
				pass = argsplit(args, input);
				if(pass.size() < 3)
				{
					litecho("\r\nToo short.");
				byebye:
					litecho(" Byebye.\r\n\n");
					close(owner.sock);
					return;
				}
				litecho("\r\nRepeat: ");
				situation = S_PASSCHECK;
				break;
			case S_PASSCHECK:
			{
				doecho();
				string s = argsplit(args, input);
				if(s != pass)
				{
				mismatch:
					litecho("\r\nMismatch.");
					goto byebye;
				}
				
				litecho(
					"\n"
					"`FYou now exist.\n\n");
				
				load("_");
				save();
				displayprompt();
				
				taskmanager += informevent(name, "has been created");
				
				break;
			}				
			case S_LOGINPASS:
			{
				doecho();
				string s = argsplit(args, input);
				if(s != pass)goto mismatch;
				litecho("\n\n");
				load();
				displayprompt();
				
				taskmanager += informevent(name, "enters the game");
				
				break;
			}				
			default:
				displayprompt();
		}
	}
	
	void takecareof(const event &ev)
	{
		unsigned char saveattr = attr;
		switch(ev.what())
		{
			case E_GLOBAL:
				owner.echo(ev.getdata(""));
				break;
			case E_INFORM:
				litecho("`Einfo: `F" + ev.getdata("name") + "`E " + ev.getdata("msg") + "\n");
				break;
			default:
				;
		}
		colour(saveattr);
	}
	
	void tick()
	{
	}
};


playertask::playertask(int s) : sock(s), lineterm(0), got_iac(0)
{
	data = new playerdata(*this);
	
	data->telnetsend(
		IAC, DO, TELOPT_BINARY,
		-1);
	
	cout << "[" << sock << "]\n";
	
	data->caninput();
}

playertask::~playertask()
{
	cout << "[~" << sock << "]\n";
	
	delete data;
	close(sock);
}

void playertask::echo(const string &s, bool endcr) const
{
	fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) &~ O_NONBLOCK);
	
	const char *q = s.c_str();
	unsigned c = s.size();
	
	while(c)
	{
		int n = write(sock, q, c);
		if(n < 0)break;
		q += n;
		c -= n;			
	}
	
	if(endcr)
		echo("\n", false);	
	else
		fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
}

bool playertask::tick()
{
	data->tick();
	
	unsigned char ch;
	int n;
	while((n = read(sock, &ch, 1)) == 1)
	{
		switch(got_iac)
		{
			case 0:
			{
				if(ch == IAC)
					++got_iac;
				else
				{
            		switch(ch)
            		{
            			case '':
            			case '':
            				if(input.size() > 0)input = input.substr(0, input.size()-1);
            				break;
            			case '\r':
            			case '\n':
            				if(lineterm && lineterm!=ch)break;
            				lineterm = ch;
            				
							cout << "[" << sock << "]: '" << input << "'\n";
							
            				data->handleinput();
            				input = "";
            				data->caninput();
            				break;
            			default:
            				if(ch >= ' ')input += ch;
            		}
            	}
				break;
			}
			case 1: // got_iac == 1
				switch(ch)
				{
					case DO:
					case DONT:
					case WILL:
					case WONT:
						got_iac = ch;
						break;
					case IP:
						input = "";
						// fall-through
						echo("^C");
					case AYT:
						// fall-through
					default:
						cout << "Got: IAC " << telcmds[ch-xEOF] << endl;
						got_iac = 0;
						break;
				}
				break;
			default:
				switch(ch)
				{
					default:
					//	if(got_iac == DO)  data->telnetsend(IAC, WILL, ch, -1);
					//	if(got_iac == DONT)data->telnetsend(IAC, WONT, ch, -1);
						cout << "Got: IAC " << telcmds[got_iac-xEOF] << " " << telopts[ch] << endl;
				}
				got_iac = 0;
		}
	}
	if(!n)return false; /* Socket died */
	if(n < 0 && errno != EAGAIN && errno != EINTR)
	{
		extern const char *sys_errlist[];
		cerr << "[!" << sock << "]: " << sys_errlist[errno] << endl;
		return false;
	}
	return true;
}

void playertask::operator += (const event &ev)
{
	data->takecareof(ev);
}
